"""
Tests for relationship module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import unittest
import numpy as np
import uncertainties
import scipy.stats
from uncertainties import unumpy as unp
from numpy.testing import assert_almost_equal, assert_equal
from uravu import UREG, utils
from uravu.relationship import Relationship
from uravu.distribution import Distribution


STRING_A = (
    "Function Name: straight_line \n"
    "Abscissa: [ 1.00e+00 2.00e+00 ... 9.00e+00 1.00e+01 ] \n"
    "Ordinate: [ 1.62e+00 -6.12e-01 ... 3.19e-01 -2.49e-01 ] \n"
    "Ordinate uncertainty: [ 1.00e-01 1.00e-01 ... 1.00e-01 1.00e-01 ]\n"
    "Abscissa Name: x \nOrdinate Name: y \nAbscissa Unit: dimensionless \n"
    "Ordinate Unit: dimensionless \nVariables: [ 1.00e+00 1.00e+00 ] \n"
    "Unaccounted uncertainty: False \nMCMC performed: False \n"
    "Nested sampling performed: False \n"
)

STRING_B = (
    "Function Name: straight_line \n"
    "Abscissa: [ 1.00e+00 5.50e+00 1.00e+01 ] \n"
    "Ordinate: [ 1.62e+00 -6.12e-01 -5.28e-01 ] \n"
    "Ordinate uncertainty: [ 1.00e-01 1.00e-01 1.00e-01 ] \n"
    "Abscissa Name: x \nOrdinate Name: y \nAbscissa Unit: dimensionless \n"
    "Ordinate Unit: dimensionless \nVariables: [ 1.00e+00 1.00e+00 ] \n"
    "Unaccounted uncertainty: False \nMCMC performed: False \n"
    "Nested sampling performed: False \n"
)

STRING_C = (
    "Function Name: straight_line \nAbscissa: [ 1.00e+00 2.00e+00 ... "
    "9.00e+00 1.00e+01 ] \nOrdinate: [ 1.62e+00 -6.12e-01 ... 3.19e-01 "
    "-2.49e-01 ] \nOrdinate uncertainty: [ 1.00e-01 1.00e-01 ... 1.00e-01 "
    "1.00e-01 ]\nAbscissa Name: x \nOrdinate Name: y \nAbscissa Unit: "
    "dimensionless \nOrdinate Unit: dimensionless \nVariables: [ 1.00e+00 "
    "1.00e+00 ] \nln(evidence): (-7.84+/-0.01)e+02 \nUnaccounted "
    "uncertainty: False \nMCMC performed: False \nNested sampling performed: "
    "True \n"
)

STRING_D = (
    "Function Name: straight_line \n"
    "Abscissa: [ 1.00e+00 5.50e+00 1.00e+01 ] \n"
    "Abscissa uncertainty: [ 1.00e-01 1.00e-01 1.00e-01 ] \n"
    "Ordinate: [ 1.62e+00 -6.12e-01 -5.28e-01 ] \n"
    "Ordinate uncertainty: [ 1.00e-01 1.00e-01 1.00e-01 ] \n"
    "Abscissa Name: x \nOrdinate Name: y \nAbscissa Unit: dimensionless \n"
    "Ordinate Unit: dimensionless \nVariables: [ 1.00e+00 1.00e+00 ] \n"
    "Unaccounted uncertainty: False \nMCMC performed: False \n"
    "Nested sampling performed: False \n"
)

STRING_E = (
    "Function Name: straight_line \nAbscissa: [ 1.00e+00 2.12e+00 ... "
    "8.88e+00 1.00e+01 ] \nAbscissa uncertainty: [ 1.00e-01 1.00e-01 ... "
    "1.00e-01 1.00e-01 ] \nOrdinate: [ 1.62e+00 -6.12e-01 ... -7.61e-01 "
    "3.19e-01 ] \nOrdinate uncertainty: [ 1.00e-01 1.00e-01 ... 1.00e-01 "
    "1.00e-01 ]\nAbscissa Name: x \nOrdinate Name: y \nAbscissa Unit: "
    "dimensionless \nOrdinate Unit: dimensionless \nVariables: [ 1.00e+00 "
    "1.00e+00 ] \nUnaccounted uncertainty: False \nMCMC performed: False "
    "\nNested sampling performed: False \n"
)

STRING_F = (
    "Function Name: straight_line \nAbscissa: [ 1.00e+00 2.00e+00 ... "
    "9.00e+00 1.00e+01 ] \nOrdinate: [ 1.62e+00 -6.12e-01 ... 3.19e-01 "
    "-2.49e-01 ] \nOrdinate uncertainty: [ 1.00e-01 1.00e-01 ... 1.00e-01 "
    "1.00e-01 ]\nAbscissa Name: x \nOrdinate Name: y \nAbscissa Unit: "
    "dimensionless \nOrdinate Unit: dimensionless \nVariables: [ "
    "-1.16e-02+5.46e+00-2.46e+00 4.43e-01+6.47e+00-6.59e+00 ] \nUnaccounted "
    "uncertainty: False \nMCMC performed: True \nNested sampling performed: "
    "False \n"
)

STRING_G = (
    "Function Name: straight_line \nAbscissa: [ 1.00e+00 2.00e+00 ... "
    "9.00e+00 1.00e+01 ] \nOrdinate: [ 1.62e+00 -6.12e-01 ... 3.19e-01 "
    "-2.49e-01 ] \nOrdinate uncertainty: [ 1.00e-01 1.00e-01 ... 1.00e-01 "
    "1.00e-01 ]\nAbscissa Name: x \nOrdinate Name: y \nAbscissa Unit: "
    "dimensionless \nOrdinate Unit: dimensionless \nVariables: [ "
    "7.01e-02+/-2.10e+00 2.31e-02+/-1.84e+00 ] \nUnaccounted "
    "uncertainty: False \nMCMC performed: True \nNested sampling performed: "
    "False \n"
)

CITATION_A = (
    "Please consider citing the following:\n"
    " - Publication of uravu (to come).\n"
    " - Zenodo DOI for uravu version: 0.0.1\n"
)
CITATION_B = (
    "Please consider citing the following:\n "
    "- Publication of uravu (to come).\n - Zenodo "
    "DOI for uravu version: 0.0.1\nThe "
    "scipy.optimize.minimize function was used to "
    "maximise the ln likelihood. Please consider citing:\n - "
    "P. Virtanen, R. Gommers, T. E. Oliphant, M. Haberland, T. "
    "Reddy, D. Cournapeau, E. Burovski, P. Peterson, W. Weckesser, "
    "J. Bright, S. J. van der Walt, M. Brett, J. Wilson, K. J. "
    "Millman, N. Mayorov, A. R. J. Nelson, E. Jones, R. Kern, "
    "E. Larson, C. Carey, I. Polat, Y. Feng, E. W. Moore, J. "
    "VanderPlas, D. Laxalde, J. Perktold, R. Cimrman, I. "
    "Henriksen, E. A. Quintero, C. R Harris, A. M. Archibald, "
    "A. H. Ribeiro, F. Pedregosa, P. van Mulbregt, & SciPy 1.0 "
    "Contributors, (2020). Nature Methods, in press. DOI: "
    "10.1038/s41592-019-0686-2\n"
)
CITATION_C = (
    "Please consider citing the following:\n - "
    "Publication of uravu (to come).\n - Zenodo DOI for uravu "
    "version: 0.0.1\nThe scipy.optimize.minimize function was used "
    "to maximise the ln likelihood. Please consider citing:\n - "
    "P. Virtanen, R. Gommers, T. E. Oliphant, M. Haberland, T. "
    "Reddy, D. Cournapeau, E. Burovski, P. Peterson, W. "
    "Weckesser, J. Bright, S. J. van der Walt, M. Brett, J. "
    "Wilson, K. J. Millman, N. Mayorov, A. R. J. Nelson, E. "
    "Jones, R. Kern, E. Larson, C. Carey, I. Polat, Y. Feng, E. W. "
    "Moore, J. VanderPlas, D. Laxalde, J. Perktold, R. Cimrman, I. "
    "Henriksen, E. A. Quintero, C. R Harris, A. M. Archibald, A. "
    "H. Ribeiro, F. Pedregosa, P. van Mulbregt, & SciPy 1.0 "
    "Contributors, (2020). Nature Methods, in press. DOI: "
    "10.1038/s41592-019-0686-2\nThe emcee package was used to "
    "perform the MCMC analysis. Please consider citing:\n - D. "
    "Foreman-Mackey, W. Farr, M. Sinha, A. Archibald, D. Hogg, J. "
    "Sanders, J. Zuntz, P. Williams, A. Nelson, M. de Val-Borro, T. "
    "Erhardt, I. Pashchenko, & O. Pla, (2019). Journal of Open "
    "Source Software, 4(43), 1864. DOI: 10.21105/joss.01864\n - "
    "J. Goodman & J. Weare, (2010). Communications in applied "
    "mathematics and computational science, 5(1), 65. DOI: "
    "10.2140/camcos.2010.5.65\n"
)
CITATION_D = (
    "Please consider citing the following:\n - Publication "
    "of uravu (to come).\n - Zenodo DOI for uravu version: "
    "0.0.1\nThe scipy.optimize.minimize function was used to "
    "maximise the ln likelihood. Please consider citing:\n - P. "
    "Virtanen, R. Gommers, T. E. Oliphant, M. Haberland, T. Reddy, "
    "D. Cournapeau, E. Burovski, P. Peterson, W. Weckesser, J. "
    "Bright, S. J. van der Walt, M. Brett, J. Wilson, K. J. "
    "Millman, N. Mayorov, A. R. J. Nelson, E. Jones, R. Kern, E. "
    "Larson, C. Carey, I. Polat, Y. Feng, E. W. Moore, J. "
    "VanderPlas, D. Laxalde, J. Perktold, R. Cimrman, I. Henriksen, "
    "E. A. Quintero, C. R Harris, A. M. Archibald, A. H. Ribeiro, "
    "F. Pedregosa, P. van Mulbregt, & SciPy 1.0 Contributors, "
    "(2020). Nature Methods, in press. DOI: "
    "10.1038/s41592-019-0686-2\nThe dynesty package was used to "
    "carry out the nested sampling. Please consider citing:\n - "
    "J. S. Speagle, (2019), Monthly Notices of the Royal "
    "Astronomical Society, staa278. DOI: "
    "10.1093/mnras/staa278.\n - J. Skilling (2004), AIP Conference "
    "Proceedings, 735(1), 395. DOI: 10.1063/1.1835238.\n - J. "
    "Skilling (2006), Bayesian Analysis, 1(4), 833. DOI: "
    "10.1214/06-BA127.\n"
)


class TestRelationship(unittest.TestCase):
    """
    Tests for the relationship module and class.
    """

    def test_init_all_default_one_dimensional(self):
        """
        Test the initialisation of the relationship class with one
        dimensional data and all defaults.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_equal(test_rel.function, utils.straight_line)
        assert_almost_equal(test_rel.abscissa.m, test_x)
        assert_almost_equal(unp.nominal_values(test_rel.ordinate.m), test_y)
        assert_almost_equal(unp.std_devs(test_rel.ordinate.m), test_y_e)
        assert_almost_equal(test_rel.variables, np.ones((2)))

    def test_init_all_default_two_dimensional(self):
        """
        Test the initialisation of the relationship class with two
        dimensional data and all defaults.
        """
        test_x = np.array([np.linspace(0, 99, 100), np.linspace(0, 99, 100)]).T
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_equal(test_rel.function, utils.straight_line)
        assert_almost_equal(test_rel.abscissa.m, test_x)
        assert_almost_equal(unp.nominal_values(test_rel.ordinate.m), test_y)
        assert_almost_equal(unp.std_devs(test_rel.ordinate.m), test_y_e)
        assert_almost_equal(test_rel.variables, np.ones((2)))

    def test_init_noordinate_uncertainty_one_dimensional(self):
        """
        Test the initialisation of the relationship class with one
        dimensional data and all defaults.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_rel = Relationship(utils.straight_line, test_x, test_y)
        assert_equal(test_rel.function, utils.straight_line)
        assert_almost_equal(test_rel.abscissa.m, test_x)
        assert_almost_equal(test_rel.ordinate.m, test_y)
        assert_almost_equal(test_rel.variables, np.ones((3)))

    def test_init_noordinate_uncertainty_two_dimensional(self):
        """
        Test the initialisation of the relationship class with two
        dimensional data and all defaults.
        """
        test_x = np.array([np.linspace(0, 99, 100), np.linspace(0, 99, 100)]).T
        test_y = np.linspace(1, 199, 100)
        test_rel = Relationship(utils.straight_line, test_x, test_y)
        assert_equal(test_rel.function, utils.straight_line)
        assert_almost_equal(test_rel.abscissa.m, test_x)
        assert_almost_equal(test_rel.ordinate.m, test_y)
        assert_almost_equal(test_rel.variables, np.ones((3)))

    def test_init_additional_uncertainty_one_dimensional(self):
        """
        Test the initialisation of the relationship class with one
        dimensional data and an additional uncertainty.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(
            utils.straight_line,
            test_x,
            test_y,
            test_y_e,
            unaccounted_uncertainty=True,
        )
        assert_equal(test_rel.function, utils.straight_line)
        assert_almost_equal(test_rel.abscissa.m, test_x)
        assert_almost_equal(unp.nominal_values(test_rel.ordinate.m), test_y)
        assert_almost_equal(unp.std_devs(test_rel.ordinate.m), test_y_e)
        assert_almost_equal(test_rel.variables, np.ones((3)))

    def test_init_additional_uncertainty_two_dimensional(self):
        """
        Test the initialisation of the relationship class with two
        dimensional data and an additional uncertainty.
        """
        test_x = np.array([np.linspace(0, 99, 100), np.linspace(0, 99, 100)]).T
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(
            utils.straight_line,
            test_x,
            test_y,
            test_y_e,
            unaccounted_uncertainty=True,
        )
        assert_equal(test_rel.function, utils.straight_line)
        assert_almost_equal(test_rel.abscissa.m, test_x)
        assert_almost_equal(unp.nominal_values(test_rel.ordinate.m), test_y)
        assert_almost_equal(unp.std_devs(test_rel.ordinate.m), test_y_e)
        assert_almost_equal(test_rel.variables, np.ones((3)))

    def test_init_different_length_x_and_y_one_dimension(self):
        """
        Test initialisation with different array lengths and one dimensional
        data.
        """
        with self.assertRaises(ValueError):
            test_x = np.linspace(0, 99, 100)
            test_y = np.linspace(1, 199, 99)
            test_y_e = test_y * 0.1
            Relationship(utils.straight_line, test_x, test_y, test_y_e)

    def test_init_different_length_x_and_y_two_dimension(self):
        """
        Test initialisation with different array lengths and two dimensional
        data.
        """
        with self.assertRaises(ValueError):
            test_x = np.array(
                [np.linspace(0, 99, 100), np.linspace(0, 99, 100)]
            ).T
            test_y = np.linspace(1, 199, 99)
            test_y_e = test_y * 0.1
            Relationship(utils.straight_line, test_x, test_y, test_y_e)

    def test_init_different_length_y_and_y_err_one_dimension(self):
        """
        Test initialisation with different array lengths and one dimensional
        data.
        """
        with self.assertRaises(ValueError):
            test_x = np.linspace(0, 99, 100)
            test_y = np.linspace(1, 199, 100)
            test_y_e = np.linspace(1, 199, 99)
            Relationship(utils.straight_line, test_x, test_y, test_y_e)

    def test_init_different_length_y_and_y_err_two_dimension(self):
        """
        Test initialisation with different array lengths and one dimensional
        data.
        """
        with self.assertRaises(ValueError):
            test_x = np.array(
                [np.linspace(0, 99, 100), np.linspace(0, 99, 100)]
            ).T
            test_y = np.linspace(1, 199, 100)
            test_y_e = np.linspace(1, 199, 99)
            Relationship(utils.straight_line, test_x, test_y, test_y_e)

    def test_init_different_variables_and_names(self):
        """
        Test initialisation with different length variable_names and
        variables
        """
        with self.assertRaises(ValueError):
            test_x = np.linspace(0, 99, 100)
            test_y = np.linspace(1, 199, 100)
            test_y_e = np.linspace(1, 199, 99)
            Relationship(
                utils.straight_line,
                test_x,
                test_y,
                test_y_e,
                variable_names=["a"],
            )

    def test_init_different_variables_and_units(self):
        """
        Test initialisation with different length variable_units and
        variables
        """
        with self.assertRaises(ValueError):
            test_x = np.linspace(0, 99, 100)
            test_y = np.linspace(1, 199, 100)
            test_y_e = np.linspace(1, 199, 99)
            Relationship(
                utils.straight_line,
                test_x,
                test_y,
                test_y_e,
                variable_units=[UREG.meter],
            )

    def test_init_variable_units(self):
        """
        Test the initialisation of the relationship class with one
        dimensional data and variable units.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(
            utils.straight_line,
            test_x,
            test_y,
            test_y_e,
            variable_units=[UREG.meter, UREG.second],
        )
        assert_equal(test_rel.function, utils.straight_line)
        assert_almost_equal(test_rel.abscissa.m, test_x)
        assert_almost_equal(unp.nominal_values(test_rel.ordinate.m), test_y)
        assert_almost_equal(unp.std_devs(test_rel.ordinate.m), test_y_e)
        assert_almost_equal(test_rel.variables, np.ones((2)))
        assert_equal(test_rel.variable_units, [UREG.meter, UREG.second])

    def test_init_variable_names(self):
        """
        Test the initialisation of the relationship class with one
        dimensional data and variable names.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(
            utils.straight_line,
            test_x,
            test_y,
            test_y_e,
            variable_names=["g", "a"],
        )
        assert_equal(test_rel.function, utils.straight_line)
        assert_almost_equal(test_rel.abscissa.m, test_x)
        assert_almost_equal(unp.nominal_values(test_rel.ordinate.m), test_y)
        assert_almost_equal(unp.std_devs(test_rel.ordinate.m), test_y_e)
        assert_almost_equal(test_rel.variables, np.ones((2)))
        assert_equal(test_rel.variable_names, ["g", "a"])

    def test_init_x_unit_one_dimensional(self):
        """
        Test the initialisation of the relationship class with one
        dimensional data and an unit in abscissa.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(
            utils.straight_line,
            test_x,
            test_y,
            test_y_e,
            abscissa_unit=UREG.meter,
        )
        assert_equal(test_rel.function, utils.straight_line)
        assert_almost_equal(test_rel.abscissa.m, test_x)
        assert_equal(test_rel.abscissa.u, UREG.meter)
        assert_almost_equal(unp.nominal_values(test_rel.ordinate.m), test_y)
        assert_almost_equal(unp.std_devs(test_rel.ordinate.m), test_y_e)
        assert_almost_equal(test_rel.variables, np.ones((2)))

    def test_init_y_unit_one_dimensional(self):
        """
        Test the initialisation of the relationship class with one
        dimensional data and an unit in ordinate.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(
            utils.straight_line,
            test_x,
            test_y,
            test_y_e,
            ordinate_unit=UREG.meter,
        )
        assert_equal(test_rel.function, utils.straight_line)
        assert_almost_equal(test_rel.abscissa.m, test_x)
        assert_equal(test_rel.ordinate.u, UREG.meter)
        assert_almost_equal(unp.nominal_values(test_rel.ordinate.m), test_y)
        assert_almost_equal(unp.std_devs(test_rel.ordinate.m), test_y_e)
        assert_almost_equal(test_rel.variables, np.ones((2)))

    def test_x_a(self):
        """
        Test the x property with no uncertainty.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_almost_equal(test_rel.x.m, test_x)
        assert_equal(test_rel.x.u, UREG.dimensionless)

    def test_y(self):
        """
        Test the y property.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_almost_equal(unp.nominal_values(test_rel.y.m), test_y)
        assert_almost_equal(unp.std_devs(test_rel.y.m), test_y_e)
        assert_equal(test_rel.y.u, UREG.dimensionless)

    def test_x_m_a(self):
        """
        Test the x_m property with no uncertainty.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_almost_equal(test_rel.x_m, test_x)

    def test_y_m(self):
        """
        Test the y_m property.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_almost_equal(unp.nominal_values(test_rel.y_m), test_y)
        assert_almost_equal(unp.std_devs(test_rel.y_m), test_y_e)

    def test_x_u(self):
        """
        Test the x_u property.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(
            utils.straight_line,
            test_x,
            test_y,
            test_y_e,
            abscissa_unit=UREG.meter,
        )
        assert_equal(test_rel.x_u, UREG.meter)

    def test_y_u(self):
        """
        Test the y_u property.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(
            utils.straight_line,
            test_x,
            test_y,
            test_y_e,
            ordinate_unit=UREG.meter,
        )
        assert_equal(test_rel.y_u, UREG.meter)

    def test_x_n_a(self):
        """
        Test the x_n property with no uncertainty.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_almost_equal(test_rel.x_n, test_x)

    def test_y_n(self):
        """
        Test the y_n property.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_almost_equal(test_rel.y_n, test_y)

    def test_y_s_a(self):
        """
        Test the y_s property.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_almost_equal(test_rel.y_s, test_y_e)

    def test_y_s_b(self):
        """
        Test the y_s property when there is no uncertainty.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_rel = Relationship(utils.straight_line, test_x, test_y)
        assert_equal(test_rel.y_s, None)

    def test_variable_medians_b(self):
        """
        Test variable_medians property when the variables are Distributions.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = (
            np.linspace(1, 199, 10)
            + np.linspace(1, 199, 10) * np.random.randn(10) * 0.05
        )
        test_y_e = test_y * 0.2
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e,)
        test_rel.max_likelihood()
        test_rel.mcmc(n_burn=10, n_samples=10)
        medians = test_rel.variable_medians
        assert_equal(medians.shape, (2,))

    def test_variable_medians_a(self):
        """
        Test variable_medians property when the variables are floats.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = (
            np.linspace(1, 199, 10)
            + np.linspace(1, 199, 10) * np.random.randn(10) * 0.05
        )
        test_y_e = test_y * 0.2
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e,)
        test_rel.max_likelihood()
        medians = test_rel.variable_medians
        assert_equal(medians.shape, (2,))

    def test_len_parameters(self):
        """
        test the len_parameters function.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_equal(test_rel.len_parameters(), 2)

    def test_max_likelihood(self):
        """
        Test max_likelihood function.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        test_rel.max_likelihood()
        assert_almost_equal(test_rel.variables, np.array([2, 1]))

    def test_prior_a(self):
        """
        Test prior function.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        test_rel.max_likelihood()
        result_priors = test_rel.prior()
        assert_equal(len(result_priors), 2)
        assert_equal(
            isinstance(
                result_priors[0], scipy.stats._distn_infrastructure.rv_frozen
            ),
            True,
        )
        assert_equal(
            isinstance(
                result_priors[1], scipy.stats._distn_infrastructure.rv_frozen
            ),
            True,
        )

    def test_prior_b(self):
        """
        Test prior function with unaccounted uncertainty.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = Relationship(
            utils.straight_line,
            test_x,
            test_y,
            test_y_e,
            unaccounted_uncertainty=True,
        )
        test_rel.max_likelihood()
        result_priors = test_rel.prior()
        assert_equal(len(result_priors), 3)
        assert_equal(
            isinstance(
                result_priors[0], scipy.stats._distn_infrastructure.rv_frozen
            ),
            True,
        )
        assert_equal(
            isinstance(
                result_priors[1], scipy.stats._distn_infrastructure.rv_frozen
            ),
            True,
        )
        assert_equal(
            isinstance(
                result_priors[2], scipy.stats._distn_infrastructure.rv_frozen
            ),
            True,
        )

    def test_mcmc(self):
        """
        Test mcmc function.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = (
            np.linspace(1, 199, 10)
            + np.linspace(1, 199, 10) * np.random.randn(10) * 0.05
        )
        test_y_e = test_y * 0.2
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e,)
        test_rel.max_likelihood()
        test_rel.mcmc(n_burn=10, n_samples=10)
        assert_equal(isinstance(test_rel.variables[0], Distribution), True)
        assert_equal(isinstance(test_rel.variables[1], Distribution), True)
        assert_equal(test_rel.variables[0].size, 1000)
        assert_equal(test_rel.variables[1].size, 1000)
        assert_equal(test_rel.mcmc_done, True)
        assert_equal(test_rel.nested_sampling_done, False)

    def test_nested_sampling(self):
        """
        Test nested sampling.
        """
        test_y = np.ones(10) * np.random.randn(10)
        test_y_e = np.ones(10) * 0.1
        test_x = np.linspace(1, 10, 10)
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        test_rel.nested_sampling(maxiter=10)
        assert_equal(
            isinstance(test_rel.ln_evidence, uncertainties.core.Variable), True
        )
        assert_equal(test_rel.nested_sampling_done, True)
        assert_equal(test_rel.mcmc_done, False)

    def test_str_a(self):
        """
        Test __str__ function A.
        """
        np.random.seed(1)
        test_y = np.ones(10) * np.random.randn(10)
        test_y_e = np.ones(10) * 0.1
        test_x = np.linspace(1, 10, 10)
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_equal(test_rel.__str__(), STRING_A)

    def test_str_b(self):
        """
        Test __str__ function B.
        """
        np.random.seed(1)
        test_y = np.ones(3) * np.random.randn(3)
        test_y_e = np.ones(3) * 0.1
        test_x = np.linspace(1, 10, 3)
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_equal(test_rel.__str__(), STRING_B)

    def test_str_c(self):
        """
        Test __str__ function C.
        """
        np.random.seed(1)
        test_y = np.ones(10) * np.random.randn(10)
        test_y_e = np.ones(10) * 0.1
        test_x = np.linspace(1, 10, 10)
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        test_rel.nested_sampling(maxiter=10)
        assert_equal(test_rel.__str__(), STRING_C)

    def test_str_f(self):
        """
        Test __str__ function F.
        """
        np.random.seed(1)
        test_y = np.ones(10) * np.random.randn(10)
        test_y_e = np.ones(10) * 0.1
        test_x = np.linspace(1, 10, 10)
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        test_rel.mcmc(n_burn=10, n_samples=10)
        assert_equal(test_rel.__str__(), STRING_F)

    def test_str_g(self):
        """
        Test __str__ function G.
        """
        np.random.seed(1)
        test_y = np.ones(10) * np.random.randn(10)
        test_y_e = np.ones(10) * 0.1
        test_x = np.linspace(1, 10, 10)
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        test_rel.mcmc(n_burn=10, n_samples=10)
        for i in range(len(test_rel.variables)):
            test_rel.variables[i] = Distribution(np.random.randn((1000)))
        assert_equal(test_rel.__str__(), STRING_G)

    def test_repr(self):
        """
        Test __repr__ function.
        """
        np.random.seed(1)
        test_y = np.ones(10) * np.random.randn(10)
        test_y_e = np.ones(10) * 0.1
        test_x = np.linspace(1, 10, 10)
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_equal(test_rel.__repr__(), STRING_A)

    def test_bic(self):
        """
        Test bayesian_information_criteria function.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = np.ones(10)
        test_y_e = np.ones(10) * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        expected_bic = -41.44653167325224
        actual_bic = test_rel.bayesian_information_criteria()
        assert_almost_equal(actual_bic, expected_bic)

    def test_citations_a(self):
        """
        test citations a.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = np.ones(10)
        test_y_e = np.ones(10) * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        assert_equal(test_rel.citations, CITATION_A)

    def test_citations_b(self):
        """
        test citations b.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = np.ones(10) * 2
        test_y_e = np.ones(10) * 0.1
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        test_rel.max_likelihood()
        assert_equal(test_rel.citations, CITATION_B)

    def test_citations_c(self):
        """
        test citations c.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = (
            np.linspace(1, 199, 10)
            + np.linspace(1, 199, 10) * np.random.randn(10) * 0.05
        )
        test_y_e = test_y * 0.2
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e,)
        test_rel.max_likelihood()
        test_rel.mcmc(n_burn=10, n_samples=10)
        assert_equal(test_rel.citations, CITATION_C)

    def test_citations_d(self):
        """
        test citations d.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = (
            np.linspace(1, 199, 10)
            + np.linspace(1, 199, 10) * np.random.randn(10) * 0.05
        )
        test_y_e = test_y * 0.2
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e,)
        test_rel.max_likelihood()
        test_rel.nested_sampling(maxiter=10)
        assert_equal(test_rel.citations, CITATION_D)



"""
Tests for distribution module

Copyright (c) Andrew R. McCluskey

Distributed under the terms of the MIT License

@author: Andrew R. McCluskey
"""

# pylint: disable=R0201

import unittest
import numpy as np
from numpy.testing import assert_almost_equal, assert_equal
from uravu.distribution import Distribution


EXPECTED_STRING1 = (
    "Distribution: Distribution\n"
    "Size: 4\n"
    "Samples: [1.62e+00 -6.12e-01 ... -5.28e-01 -1.07e+00]\n"
    "Median: -5.70e-01\n"
    "Confidence intervals: [-1.04e+00 1.46e+00]\n"
    "Confidence interval points: [2.5 97.5]\n"
    "Reporting Value: -5.70e-01+2.03e+00-4.68e-01\n"
    "Unit: dimensionless\n"
    "Normal: False\n"
)

EXPECTED_STRING2 = (
    "Distribution: Distribution\nSize: 1000\nSamples: [1.62e+00 -6.12e-01 "
    "... 3.54e-01 -1.87e-01]\nMedian: 4.13e-02\nSymetrical Error: "
    "9.81e-01\nConfidence intervals: [-1.96e+00 1.91e+00]\nConfidence "
    "interval points: [2.5 97.5]\nReporting Value: 0.0+/-1.0\nUnit: "
    "dimensionless\nNormal: True\n"
)


class TestDistribution(unittest.TestCase):
    """
    Testing the Distribution class.
    """

    def test_init_a(self):
        """
        Test initialisation with defaults.
        """
        distro = Distribution([1])
        assert_equal(distro.samples, np.array([1]))
        assert_equal(distro.n, 1)
        assert_almost_equal(distro.con_int, [1, 1])
        assert_almost_equal(distro.ci_points, [2.5, 97.5])

    def test_init_b(self):
        """
        Test initialisation without defaults.
        """
        distro = Distribution([1], ci_points=[5.0, 95.0])
        assert_equal(distro.samples, np.array([1]))
        assert_equal(distro.n, 1)
        assert_almost_equal(distro.con_int, [1, 1])
        assert_almost_equal(distro.ci_points, [5.0, 95.0])

    def test_init_c(self):
        """
        Test initialisation with bad ci_points.
        """
        with self.assertRaises(ValueError):
            Distribution([1, 2], ci_points=[5.0, 95.0, 102.0])

    def test_check_normality_true_less_than_5000(self):
        """
        Test check_normality with less than 5000 samples.
        """
        np.random.seed(1)
        distro = Distribution(np.random.randn(4999))
        assert_equal(distro.normal, True)

    def test_check_normality_true_more_than_5000(self):
        """
        Test check_normality with more than 5000 samples.
        """
        np.random.seed(1)
        distro = Distribution(np.random.randn(10000))
        assert_equal(distro.normal, True)

    def test_check_normality_false_more_than_5000(self):
        """
        Test check_normality with more than 5000 samples.
        """
        np.random.seed(1)
        distro = Distribution(np.random.rand(10000))
        assert_equal(distro.normal, False)

    def test_check_normality_false_less_than_5000(self):
        """
        Test check_normality with less than 5000 samples.
        """
        np.random.seed(1)
        distro = Distribution(np.random.rand(1000))
        assert_equal(distro.normal, False)

    def test_check_normality_less_than_3(self):
        """
        Test check_normality with more than 5000 samples.
        """
        np.random.seed(1)
        distro = Distribution(np.random.randn(2))
        assert_equal(distro.normal, False)

    def test_mean_a(self):
        """
        Test the mean parameter
        """
        np.random.seed(1)
        distro = Distribution(np.random.randn(2000))
        assert_almost_equal(distro.mean.n, 0.034504725958795925)
        assert_almost_equal(distro.mean.s, 1.0058360180170436)

    def test_mean_b(self):
        """
        Test the mean parameter when not normal
        """
        np.random.seed(1)
        distro = Distribution(np.random.random(2000))
        assert_equal(distro.mean, None)

    def test_add_samples_single(self):
        """
        Test add_samples with a single value.
        """
        distro = Distribution(1)
        assert_equal(distro.size, 1)
        assert_almost_equal(distro.samples, np.array([1]))
        assert_almost_equal(distro.n, 1)
        assert_equal(distro.s, None)
        assert_almost_equal(distro.con_int, [1, 1])
        assert_equal(distro.normal, False)

    def test_string_output_a(self):
        """
        Test the string that is printed.
        """
        np.random.seed(1)
        distro = Distribution(np.random.randn((1000)))
        assert_equal(distro.__str__(), EXPECTED_STRING2)

    def test_string_output_b(self):
        """
        Test the string that is printed.
        """
        np.random.seed(1)
        distro = Distribution(np.random.randn((4)))
        assert_equal(distro.__str__(), EXPECTED_STRING1)

    def test_repr_output(self):
        """
        Test the representation that is printed.
        """
        np.random.seed(1)
        distro = Distribution(np.random.randn((1000)))
        assert_equal(distro.__repr__(), EXPECTED_STRING2)



"""
These are plotting functions that take either :class:`Relationship`
or :class:`Distribution` class objects.

The aim is to produce *publication quality* plots. However, we
recognise that taste exists, and ours may be different from yours.
The colorscheme in this work was chosen to be colorblind friendly.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import numpy as np
import uncertainties
from scipy.stats import gaussian_kde
from uravu import UREG, _fig_params
from uravu.distribution import Distribution
try:
    import matplotlib.pyplot as plt
    from corner import corner
except ModuleNotFoundError:
    print("The matplotlib and corner packages are necessary for the use of "
          "the plotting module, please install these.")


def plot_relationship(
    relationship, axes=None, figsize=(10, 6)
):  # pragma: no cover
    """
    Plot the relationship. Additional plots will be included on this if
    the MCMC sampling has been used to find distributions.

    Args:
        relationship (uravu.relationship.Relationship): The relationship to
            be plotted.
        axes (matplotlib.axes): Axes to which the plot should be added.
            If none given new axes will be created.
        fig_size (tuple, optional): horizontal and veritcal size for
            figure (in inches). Default is `(10, 6)`.

    Returns:
        (matplotlib.axes): The axes with new plots.
    """
    if axes is None:
        axes = plt.subplots(figsize=figsize)[1]
    variables = relationship.variables
    if relationship.unaccounted_uncertainty:
        variables = relationship.variables[:-1]
    axes.plot(
        relationship.x_n, relationship.y_n, c=list(_fig_params.TABLEAU)[0]
    )
    x_label = "{}".format(relationship.abscissa_name)
    if relationship.x_u != UREG.dimensionless:
        x_label += "/${:~L}$".format(relationship.x_u)
    axes.set_xlabel(x_label)
    y_label = "{}".format(relationship.ordinate_name)
    if relationship.y_u != UREG.dimensionless:
        y_label += "/${:~L}$".format(relationship.y_u)
    axes.set_ylabel(y_label)
    if isinstance(
        relationship.ordinate.m.any(), uncertainties.core.AffineScalarFunc
    ):
        axes.fill_between(
            relationship.x_n,
            relationship.y_n - relationship.y_s,
            relationship.y_n + relationship.y_s,
            alpha=0.5,
            color=list(_fig_params.TABLEAU)[0],
            lw=0,
        )
    if not isinstance(variables[0], Distribution):
        axes.plot(
            relationship.x_n,
            relationship.function(relationship.x_n, *variables),
            color=list(_fig_params.TABLEAU)[1],
        )
    else:
        plot_samples = np.random.randint(
            0, variables[0].samples.size, size=100
        )
        for i in plot_samples:
            float_variables = [var.samples[i] for var in variables]
            axes.plot(
                relationship.x_n,
                relationship.function(relationship.x_n, *float_variables),
                color=list(_fig_params.TABLEAU)[1],
                alpha=0.05,
            )
    if relationship.unaccounted_uncertainty:
        var = relationship.variable_medians
        additional_uncertainty = np.abs(
            var[-1] * relationship.function(relationship.x_n, *var[:-1])
        )
        axes.fill_between(
            relationship.x_n,
            relationship.y_n + relationship.y_s,
            relationship.y_n + relationship.y_s + additional_uncertainty,
            alpha=0.5,
            color=list(_fig_params.TABLEAU)[2],
            lw=0,
        )
        axes.fill_between(
            relationship.x_n,
            relationship.y_n - relationship.y_s,
            relationship.y_n - relationship.y_s - additional_uncertainty,
            alpha=0.5,
            color=list(_fig_params.TABLEAU)[2],
            lw=0,
        )
    return axes


def plot_distribution(distro, axes=None, figsize=(10, 6)):  # pragma: no cover
    """
    Plot the probability density function for a distribution.

    Args:
        distro (uravu.distriobution.Distribution): The distribution to be
            plotted.
        fig_size (tuple): Horizontal and veritcal size for figure
            (in inches).

    Returns:
        (matplotlib.axes): The axes with new plots.
    """
    if axes is None:
        axes = plt.subplots(figsize=figsize)[1]
    kde = gaussian_kde(distro.samples)
    abscissa = np.linspace(distro.samples.min(), distro.samples.max(), 100)
    ordinate = kde.evaluate(abscissa)
    axes.plot(abscissa, ordinate, color=list(_fig_params.TABLEAU)[0])
    axes.hist(
        distro.samples,
        bins=25,
        density=True,
        color=list(_fig_params.TABLEAU)[0],
        alpha=0.5,
    )
    axes.fill_betweenx(
        np.linspace(0, ordinate.max() + ordinate.max() * 0.1),
        distro.con_int[0],
        distro.con_int[1],
        alpha=0.2,
    )
    x_label = "{}".format(distro.name)
    if distro.unit != UREG.dimensionless:
        x_label += "/${:~L}$".format(distro.unit)
    axes.set_xlabel(x_label)
    axes.set_ylabel("$p(${}$)$".format(distro.name))
    axes.set_ylim((0, ordinate.max() + ordinate.max() * 0.1))
    return axes


def plot_corner(relationship, figsize=(8, 8)):  # pragma: no cover
    """
    Plot the corner (named for the Python package) plot between the
    relationships variables.

    Args:
        relationship (uravu.relationship.Relationship): The relationship
            containing the distributions to be plotted.
        fig_size (tuple, optional): horizontal and veritcal size for
            figure (in inches). Default is `(10, 6)`.

    Returns:
        (matplotlib.figure): The figure with new plots.
        (matplotlib.axes): The axes with new plots.
    """
    n = len(relationship.variables)
    if not all(
        [isinstance(relationship.variables[i], Distribution) for i in range(n)]
    ):
        raise ValueError(
            "In order to use the corner plot functionality, all relationship "
            "variables must be Distributions. Please run MCMC before "
            "plotting the corner."
        )
    fig, ax = plt.subplots(n, n, figsize=figsize)
    var_labels = []
    for i in range(n):
        if relationship.variable_units[i] == UREG.dimensionless:
            var_labels.append("{}".format(relationship.variable_names[i]))
        else:
            var_labels.append(
                "{}/${:L}$".format(
                    relationship.variable_names[i],
                    relationship.variable_units[i],
                )
            )
    corner(
        relationship.mcmc_results["samples"],
        color=list(_fig_params.TABLEAU)[0],
        hist_kwargs={"lw": 4, "histtype": "stepfilled"},
        label_kwargs={"fontsize": _fig_params.rcParams["axes.labelsize"]},
        fig=fig,
        labels=var_labels,
    )
    for j in range(n):
        ax[n - 1, j].set_xticks(
            [
                i
                for i in np.percentile(
                    relationship.variables[j].samples, [2.5, 50, 97.5]
                )
            ]
        )
        ax[n - 1, j].set_xlim(
            [
                i
                for i in np.percentile(
                    relationship.variables[j].samples, [0.5, 99.5]
                )
            ]
        )
    for j in range(n - 1):
        ax[j + 1, 0].set_yticks(
            [
                i
                for i in np.percentile(
                    relationship.variables[j + 1].samples, [2.5, 50, 97.5]
                )
            ]
        )
        ax[j + 1, 0].set_ylim(
            [
                i
                for i in np.percentile(
                    relationship.variables[j + 1].samples, [0.5, 99.5]
                )
            ]
        )
    return fig, ax



"""
Tests for sampling module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import unittest
import numpy as np
from numpy.testing import assert_equal
from scipy.stats import norm
from uravu import sampling, utils
from uravu.relationship import Relationship
from uravu.distribution import Distribution


class TestSampling(unittest.TestCase):
    """
    Unit tests for optimize module.
    """

    def test_mcmc_a(self):
        """
        Test mcmc function.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = (
            np.linspace(1, 199, 10)
            + np.linspace(1, 199, 10) * np.random.randn(10) * 0.05
        )
        test_y_e = test_y * 0.2
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e,)
        test_rel.max_likelihood()
        actual_results = sampling.mcmc(test_rel, n_burn=10, n_samples=10)
        assert_equal(
            isinstance(actual_results["distributions"][0], Distribution), True
        )
        assert_equal(
            isinstance(actual_results["distributions"][1], Distribution), True
        )
        assert_equal(actual_results["distributions"][0].size, 1000)
        assert_equal(actual_results["distributions"][1].size, 1000)

    def test_mcmc_b(self):
        """
        Test mcmc function with custom prior.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = (
            np.linspace(1, 199, 10)
            + np.linspace(1, 199, 10) * np.random.randn(10) * 0.05
        )
        test_y_e = test_y * 0.2
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e,)
        test_rel.max_likelihood()

        def other_prior():
            """
            Another potential prior.
            """
            priors = []
            for i, variable in enumerate(test_rel.variables):
                loc = variable
                scale = 1
                priors.append(norm(loc=loc, scale=scale))
            return priors

        actual_results = sampling.mcmc(
            test_rel, prior_function=other_prior, n_burn=10, n_samples=10
        )
        assert_equal(
            isinstance(actual_results["distributions"][0], Distribution), True
        )
        assert_equal(
            isinstance(actual_results["distributions"][1], Distribution), True
        )
        assert_equal(actual_results["distributions"][0].size, 1000)
        assert_equal(actual_results["distributions"][1].size, 1000)

    def test_nested_sampling_a(self):
        """
        Test nested sampling.
        """
        test_y = np.ones(10) * np.random.randn(10)
        test_y_e = np.ones(10) * 0.1
        test_x = np.linspace(1, 10, 10)
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        actual_results = sampling.nested_sampling(test_rel, maxiter=10)
        assert_equal(isinstance(actual_results, dict), True)
        assert_equal(isinstance(actual_results["logz"][-1], float), True)
        assert_equal(isinstance(actual_results["logzerr"][-1], float), True)

    def test_nested_sampling_b(self):
        """
        Test nested_sampling function with custom prior.
        """
        test_y = np.ones(10) * np.random.randn(10)
        test_y_e = np.ones(10) * 0.1
        test_x = np.linspace(1, 10, 10)
        test_rel = Relationship(utils.straight_line, test_x, test_y, test_y_e)
        test_rel.max_likelihood()

        def other_prior():
            """
            Another potential prior.
            """
            priors = []
            for i, variable in enumerate(test_rel.variables):
                loc = variable
                scale = 1
                priors.append(norm(loc=loc, scale=scale))
            return priors

        actual_results = sampling.nested_sampling(
            test_rel, prior_function=other_prior, maxiter=10
        )
        assert_equal(isinstance(actual_results, dict), True)
        assert_equal(isinstance(actual_results["logz"][-1], float), True)
        assert_equal(isinstance(actual_results["logzerr"][-1], float), True)



"""
A few additional utility functions to improve the usability of
``uravu``.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey


import numpy as np
from scipy.stats import pearsonr


def straight_line(abscissa, gradient, intercept):
    """
    A one dimensional straight line function.

    Args:
        abscissa (array_like): the abscissa data.
        gradient (float): the slope of the line.
        intercept (float): the y-intercept of the line.

    Returns:
        (array_like): the resulting ordinate.
    """
    return gradient * abscissa + intercept


def bayes_factor(model_1, model_2):
    """
    Use the Bayes factor to compare two models.

    Args:
        model_1 (uncertainties.core.Variable or float): ln evidence for
            model 1.
        model_2 (uncertainties.core.Variable or float): ln evidence for
            model 2.

    Return:
        (uncertainties.core.Variable or float): 2ln(B) where B is the Bayes
            Factor between the two models.
    """
    return 2 * (model_1 - model_2)


def correlation_matrix(relationship):
    """
    Evalutate the Pearsons correlation coefficient matrix for the
    variables in a given relationship.

    Args:
        relationship (uravu.relationship.Relationship): The relationship
            to determine the posteriors of.

    Returns:
        (array_like): The correlation matrix for the relationships
            variables.
    """
    n = len(relationship.variables)
    matrix = np.zeros((n, n))
    subset = np.random.randint(0, relationship.variables[0].size, size=400)
    for i in range(n):
        samples_i = relationship.variables[i].samples[subset]
        for j in range(i, n):
            samples_j = relationship.variables[j].samples[subset]
            matrix[i, j] = pearsonr(samples_i, samples_j)[0]
            matrix[j, i] = pearsonr(samples_j, samples_i)[0]
    return matrix



"""
The optimize module includes the functionality necessary for maximum
likelihood determination. Further, the natural log likelihood function used
in the ``mcmc`` and ``nested_sampling`` methods may be found here.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import numpy as np
import uncertainties
from scipy.optimize import minimize
from uncertainties import unumpy as unp


def max_ln_likelihood(relationship, x0=None, **kwargs):
    """
    Determine the variable values which maximize the likelihood for the
    given relationship. For keyword arguments see the
    `scipy.optimize.minimize()`_ documentation.

    Args:
        relationship (uravu.relationship.Relationship): The relationship for
            which variables should be found.
        x0 (array_like, optional): Initial guesses for the variable values.
            Default to the current ``relationship.variables`` values which
            are initialized as all ones.

    Return:
        (array_like): optimized variables for the relationship.

    .. _scipy.optimize.minimize(): https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
    """
    if x0 is None:
        x0 = relationship.variable_medians
    return minimize(
        negative_lnl,
        x0,
        args=(
            relationship.function,
            relationship.abscissa,
            relationship.ordinate,
            relationship.unaccounted_uncertainty,
        ),
        **kwargs,
    ).x


def negative_lnl(
    variables, function, abscissa, ordinate, unaccounted_uncertainty=False
):
    """
    Calculate the negative natural logarithm of the likelihood given a set
    of variables, when there is no uncertainty in the abscissa.

    Args:
        variables (array_like): Variables for the function evaluation.
        function (callable): The function to be evaluated.
        abscissa (array_like): The abscissa values.
        ordinate (array_like): The ordinate values.
        unaccounted_uncertainty (bool, optional): Should an unaccounted
            uncertainty parameter be considered. Default is ``False``.

    Returns:
        (float): Negative ln-likelihood between model and data.
    """
    return -ln_likelihood(
        variables,
        function,
        abscissa,
        ordinate,
        unaccounted_uncertainty=unaccounted_uncertainty,
    )


def ln_likelihood(
    variables, function, abscissa, ordinate, unaccounted_uncertainty=False
):
    """
    Calculate the natural logarithm of the likelihood given a set of
    variables, when there is no uncertainty in the abscissa.

    Args:
        variables (array_like): Variables for the function evaluation.
        function (callable): The function to be evaluated.
        abscissa (array_like): The abscissa values.
        ordinate (array_like): The ordinate values.
        unaccounted_uncertainty (bool, optional): Should an unaccounted
            uncertainty parameter be considered. Default is ``False``.

    Returns:
        (float): ln-likelihood between model and data.
    """
    if unaccounted_uncertainty:
        var = variables[:-1]
        log_f = variables[-1]
    else:
        var = variables
        log_f = -np.inf
    model = function(abscissa.m, *var)
    y_data = unp.nominal_values(ordinate.m)
    dy_data = unp.std_devs(ordinate.m)

    if np.isclose(dy_data.all(), 0.0):
        sigma2 = model ** 2 * np.exp(2 * log_f)
    else:
        sigma2 = dy_data ** 2 + model ** 2 * np.exp(2 * log_f)
    return -0.5 * np.sum((model - y_data) ** 2 / sigma2 + np.log(sigma2))
        



"""
The ``Relationship`` class is a foundational component of the ``uravu``
package, and acts as an API for use of the package.
This class enables the storage of the relationship between the model and
the data.

Objects of this class offer easy methods to perform maximum likelihood
evaluation, Markiv chain Monte Carlo (MCMC) for posterior probabiltiy
determination and Bayesian evidence estimation by nested sampling.

See the `tutorials online`_ for more guidence of how to use this package.

.. _tutorials online: https://uravu.rtfd.io
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

from inspect import getfullargspec
import numpy as np
import uncertainties
from scipy.stats import uniform
from uncertainties import unumpy as unp
from uncertainties import ufloat
from uravu import UREG, optimize, sampling, __version__
from uravu.distribution import Distribution


class Relationship:
    """
    The ``Relationship`` class is the base of the ``uravu`` package, enabling
    the use of Bayesian inference for the assessment of a model's ability to
    describe some data.

    Attributes:
        function (callable): The function that is modelled.
        abscissa (array_like with pint.UnitRegistry()): The abscissa data
            that the modelling should be performed from. This includes some
            unit from the `pint`_ package.
        abscissa_name (str): A name for the abscissa data, used in the
            production of plots.
        ordinate (array_like with pint.UnitRegistry()): The ordinate
            data against with the model should be compared. This may include
            uncertainty values and some unit.
        ordinate_name (str): A name for the ordinate data, used in the
            production of plots.
        unaccounted_uncertainty (bool): This boolean describes if an
            unaccounted for uncertainty should be considered in the modelling
            process.
        ln_evidence (uncertainties.core.Variable): The natural-log of the
            Bayesian evidence for the model to the given data.

    Args:
        function (callable): The functional relationship to be modelled.
        abscissa (array_like): The abscissa data. If multi-dimensional, the
            array is expected to have the shape `(N, d)`, where `N` is the
            number of data points and `d` is the dimensionality.
        ordinate (array_like): The ordinate data. This should have a
            shape `(N,)`.
        ordinate_uncertainty (array_like, optional): The uncertainty in each
            of the ordinate data points. This should have a shape `(N,)`.
            Default to no uncertainties on the ordiante. If there is no
            ordinate uncertainty, an unaccounted uncertainty is automatically
            added. 
        abscissa_unit (pint.UnitRegistry(), optional): The unit for the
            abscissa. If `abscissa` is multi-dimensional, this should be a
            list with the units for each dimension. Default is dimensionless.
        ordinate_unit (pint.UnitRegistry(), optional): The unit for the
            ordinate. Default is dimensionless.
        abscissa_name (str, optional): A name for the abscissa. Default
            is `'x'`.
        ordinate_name (str, optional): A name for the ordinate. Default
            is `'y'`.
        variable_names (list of str, optional): Names for each of the
            variables. Default is the variable name in the ``function``
            definition.
        variable_units (pint.UnitRegistry(), optional): The units for the
            variables. Default is dimensionless.
        unaccounted_uncertainty (bool, optional): Describes if an additional
            variable be included to account for an unknown uncertainty in the
            data.

    .. _pint: https://pint.readthedocs.io/
    .. _uncertainties: https://uncertainties-python-package.readthedocs.io/
    """

    def __init__(
        self,
        function,
        abscissa,
        ordinate,
        ordinate_uncertainty=None,
        abscissa_unit=UREG.dimensionless,
        ordinate_unit=UREG.dimensionless,
        abscissa_name="x",
        ordinate_name="y",
        variable_names=None,
        variable_units=None,
        unaccounted_uncertainty=False,
    ):
        self.function = function
        self.unaccounted_uncertainty = unaccounted_uncertainty
        abscissa = np.array(abscissa)
        ordinate = np.array(ordinate)

        if ordinate_uncertainty is None:
            self.unaccounted_uncertainty = True
            self.ordinate = ordinate
        else:    
            if ordinate_uncertainty.shape != ordinate.shape:
                raise ValueError(
                        "The number of data points in the ordinate does not "
                        "match that in the ordinate uncertainty."
                    )
            else:
                ordinate_uncertainty = np.array(ordinate_uncertainty)
                self.ordinate = unp.uarray(ordinate, ordinate_uncertainty) 
        self.ordinate *= ordinate_unit

        self.abscissa = abscissa
        self.abscissa *= abscissa_unit

        if abscissa.shape[0] != ordinate.shape[0]:
            raise ValueError(
                "The number of data points in the abscissa does "
                "not match that for the ordinate."
            )

        self.variables = np.ones((self.len_parameters()))
        self.abscissa_name = abscissa_name
        self.ordinate_name = ordinate_name
        if variable_names is None:
            self.variable_names = getfullargspec(self.function).args[1:]
            if self.unaccounted_uncertainty:
                self.variable_names.append("unaccounted uncertainty")
        else:
            if len(variable_names) != self.len_parameters():
                raise ValueError(
                    "The number of variable names does not match the number "
                    "of variables."
                )
            self.variable_names = variable_names
        if variable_units is None:
            self.variable_units = [
                UREG.dimensionless for i in range(self.len_parameters())
            ]
            if self.unaccounted_uncertainty:
                self.variable_units.append(self.ordinate.u)
        else:
            if len(variable_units) != self.len_parameters():
                raise ValueError(
                    "The number of variable units does not match the number "
                    "of variables."
                )
            self.variable_units = variable_units
        self.ln_evidence = None

    def __str__(self):
        """
        A custom string function.

        Returns:
            (str): Custom string description.
        """
        string = "Function Name: {} \n".format(self.function.__name__)
        if self.abscissa.shape[0] < 4:
            string += "Abscissa: [ "
            for i in self.x_n:
                string += "{:.2e} ".format(i)
            string += "] \n"
            string += "Ordinate: [ "
            for i in self.y_n:
                string += "{:.2e} ".format(i)
            string += "] \n"
            if isinstance(
                self.ordinate.m.any(), uncertainties.core.AffineScalarFunc
            ):
                string += "Ordinate uncertainty: [ "
                for i in self.y_s:
                    string += "{:.2e} ".format(i)
                string += "] \n"
        else:
            string += (
                "Abscissa: "
                "[ {:.2e} {:.2e} ... {:.2e} {:.2e} ] \n".format(
                    *self.x_n[:2], *self.x_n[-2:]
                )
            )
            string += (
                "Ordinate: "
                "[ {:.2e} {:.2e} ... {:.2e} {:.2e} ] \n".format(
                    *self.y_n[:2], *self.y_n[-2:]
                )
            )
            if isinstance(
                self.ordinate.m.any(), uncertainties.core.AffineScalarFunc
            ):
                string += (
                    "Ordinate uncertainty: "
                    "[ {:.2e} {:.2e} ... {:.2e} {:.2e} ]\n".format(
                        *self.y_s[:2], *self.y_s[-2:]
                    )
                )
        string += "Abscissa Name: {} \n".format(self.abscissa_name)
        string += "Ordinate Name: {} \n".format(self.ordinate_name)
        string += "Abscissa Unit: {} \n".format(self.x_u)
        string += "Ordinate Unit: {} \n".format(self.y_u)
        string += "Variables: [ "
        for var in self.variables:
            if isinstance(var, Distribution):
                if var.normal:
                    string += "{:.2e}+/-{:.2e} ".format(
                        var.n, var.n - var.con_int[0]
                    )
                else:
                    string += "{:.2e}+{:.2e}-{:.2e} ".format(
                        var.n, var.con_int[1] - var.n, var.n - var.con_int[0]
                    )
            else:
                string += "{:.2e} ".format(var)
        string += "] \n"
        if self.nested_sampling_done:
            string += "ln(evidence): {:.2e} \n".format(self.ln_evidence)
        string += "Unaccounted uncertainty: {} \n".format(
            self.unaccounted_uncertainty
        )
        string += "MCMC performed: {} \n".format(self.mcmc_done)
        string += "Nested sampling performed: {} \n".format(
            self.nested_sampling_done
        )
        return string

    def __repr__(self):
        """
        A custom representation.

        Returns:
            (str): Custom string description.
        """
        return self.__str__()

    @property
    def x(self):
        """
        Abscissa values with unit and uncertainty (where present).

        Returns:
            (array_like): Abscissa values.
        """
        return self.abscissa

    @property
    def y(self):
        """
        Ordinate values with unit and uncertainty.

        Returns:
            (array_like): Ordinate values.
        """
        return self.ordinate

    @property
    def x_m(self):
        """
        Abscissa values without units.

        Returns:
            (array_like): Abscissa values without units.
        """
        return self.abscissa.m

    @property
    def y_m(self):
        """
        Ordinate values without units.

        Returns:
            (array_like): Ordinate values without units.
        """
        return self.ordinate.m

    @property
    def x_u(self):
        """
        Abscissa unit.

        Returns:
            (pint.UnitRegistry()): Abscissa values.
        """
        return self.abscissa.u

    @property
    def y_u(self):
        """
        Ordinate unit.

        Returns:
            (pint.UnitRegistry()): Ordinate unit.
        """
        return self.ordinate.u

    @property
    def x_n(self):
        """
        Abscissa nominal values.

        Returns:
            (array_like): Abscissa nominal values.
        """
        return unp.nominal_values(self.abscissa.m)

    @property
    def y_n(self):
        """
        Ordinate nominal values.

        Returns:
            (array_like): Ordinate nominal values.
        """
        return unp.nominal_values(self.ordinate.m)

    @property
    def y_s(self):
        """
        Ordinate uncertainties.

        Returns:
            (array_like): Ordinate uncertainties.
        """
        if isinstance(
            self.ordinate.m.any(), uncertainties.core.AffineScalarFunc
        ):
            return unp.std_devs(self.ordinate.m)
        else:
            return None

    @property
    def variable_medians(self):
        """
        The median values for each of the variables.

        Returns:
            (array_like): The variable medians.
        """
        means = np.zeros((len(self.variables)))
        for i, var in enumerate(self.variables):
            if isinstance(var, Distribution):
                means[i] = var.n
            if isinstance(var, float):
                means[i] = var
        return means

    @property
    def mcmc_done(self):
        """
        Has MCMC been performed, determined based on the type of the variables.

        Returns:
            (bool): Has MCMC been performed.
        """
        for var in self.variables:
            if isinstance(var, Distribution):
                return True
        return False

    @property
    def nested_sampling_done(self):
        """
        Has nested sampling been performed, determined based on if the
        ln_evidence has a value.

        Returns:
            (bool): Has nested sampling been performed.
        """
        if self.ln_evidence is not None:
            return True
        return False

    @property
    def citations(self):
        """
        Return the relevant citations.

        Returns:
            (str): The citations relevant to the analysis performed.
        """
        string = "Please consider citing the following:\n"
        string += " - Publication of uravu (to come).\n"
        string += " - Zenodo DOI for uravu version: {}\n".format(__version__())
        if not np.isclose(
            self.variable_medians, np.ones((self.len_parameters()))
        ).all():
            string += (
                "The scipy.optimize.minimize function was used to "
                "maximise the ln likelihood. Please consider citing:\n"
            )
            string += (
                " - P. Virtanen, R. Gommers, T. E. Oliphant, M. "
                "Haberland, T. Reddy, D. Cournapeau, E. Burovski, P. "
                "Peterson, W. Weckesser, J. Bright, S. J. van der "
                "Walt, M. Brett, J. Wilson, K. J. Millman, N. "
                "Mayorov, A. R. J. Nelson, E. Jones, R. Kern, E. "
                "Larson, C. Carey, I. Polat, Y. Feng, E. W. Moore, "
                "J. VanderPlas, D. Laxalde, J. Perktold, R. Cimrman, "
                "I. Henriksen, E. A. Quintero, C. R Harris, A. M. "
                "Archibald, A. H. Ribeiro, F. Pedregosa, P. van "
                "Mulbregt, & SciPy 1.0 Contributors, (2020). Nature "
                "Methods, in press. DOI: 10.1038/s41592-019-0686-2\n"
            )
        if self.mcmc_done:
            string += (
                "The emcee package was used to perform the MCMC "
                "analysis. Please consider citing:\n"
            )
            string += (
                " - D. Foreman-Mackey, W. Farr, M. Sinha, A. "
                "Archibald, D. Hogg, J. Sanders, J. Zuntz, P. "
                "Williams, A. Nelson, M. de Val-Borro, T. Erhardt, "
                "I. Pashchenko, & O. Pla, (2019). Journal of Open "
                "Source Software, 4(43), 1864. DOI: "
                "10.21105/joss.01864\n"
            )
            string += (
                " - J. Goodman & J. Weare, (2010). Communications in "
                "applied mathematics and computational science, 5(1), "
                "65. DOI: 10.2140/camcos.2010.5.65\n"
            )
        if self.nested_sampling_done:
            string += (
                "The dynesty package was used to carry out the nested "
                "sampling. Please consider citing:\n"
            )
            string += (
                " - J. S. Speagle, (2019), Monthly Notices of the "
                "Royal Astronomical Society, staa278. DOI: "
                "10.1093/mnras/staa278.\n"
            )
            string += (
                " - J. Skilling (2004), AIP Conference Proceedings, "
                "735(1), 395. DOI: 10.1063/1.1835238.\n"
            )
            string += (
                " - J. Skilling (2006), Bayesian Analysis, "
                "1(4), 833. DOI: 10.1214/06-BA127.\n"
            )
        return string

    def len_parameters(self):
        """
        Determine the number of variables in the assessment function.

        Returns:
            (int): number of variables.
        """
        # The minus one is to remove the abscissa data which is a
        # argument in the assessment function
        if self.unaccounted_uncertainty:
            return len(getfullargspec(self.function).args)
        return len(getfullargspec(self.function).args) - 1

    def bayesian_information_criteria(self):
        """
        Calculate the Bayesian information criteria for the relationship.

        Returns:
            (float): Bayesian information criteria.
        """
        self.max_likelihood()
        return np.log(
            self.x_n.size
        ) * self.len_parameters() - 2 * optimize.ln_likelihood(
            self.variable_medians,
            self.function,
            self.abscissa,
            self.ordinate,
            self.unaccounted_uncertainty,
        )

    def max_likelihood(self, x0=None, **kwargs):
        """
        Determine values for the variables which maximise the likelihood
        for the relationship. For keyword arguments see the
        `scipy.optimize.minimize()`_ documentation.

        .. _scipy.optimize.minimize(): https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
        """
        self.variables = optimize.max_ln_likelihood(self, x0, **kwargs)

    def prior(self):
        """
        *Standard priors* for the relationship. These priors are broad,
        uninformative, for normal variables running the range
        [x - x * 10, x + x * 10) (where x is the variable). For an unaccounted
        uncertainty natural log factor the range is [-10, 1).

        Returns:
            (list of scipy.stats.rv_continuous): scipy.stats functions
                describing the priors.
        """
        priors = []
        for var in self.variable_medians:
            loc = var - np.abs(var) * 10
            scale = (var + np.abs(var) * 10) - loc
            priors.append(uniform(loc=loc, scale=scale))
        if self.unaccounted_uncertainty:
            priors[-1] = uniform(loc=-10, scale=11)
        return priors

    def mcmc(
        self,
        prior_function=None,
        walkers=100,
        n_samples=500,
        n_burn=500,
        progress=True,
        seed=None
    ):
        """
        Perform MCMC to get the posterior probability distributions for
        the variables of the relationship. Note running this method will
        populate the `relationship.variables` attribute with
        `uravu.distribution.Distribution` objects.

        Args:
            prior_function (callable, optional): the function to populated
                some prior distributions. Default is the broad uniform
                priors in uravu.relationship.Relationship.
            walkers (int, optional): Number of MCMC walkers. Default is `100`.
            n_samples (int, optional): Number of sample points. Default
                is `500`.
            n_burn (int, optional): Number of burn in samples. Default
                is `500`.
            progress (bool, optional): Show tqdm progress for sampling.
                Default is `True`.
        """
        self.mcmc_results = sampling.mcmc(
            self,
            prior_function=prior_function,
            walkers=walkers,
            n_samples=n_samples,
            n_burn=n_burn,
            progress=progress,
        )
        self.variables = self.mcmc_results["distributions"]

    def nested_sampling(self, prior_function=None, progress=True, **kwargs):
        """
        Perform nested sampling to determine the Bayesian
        natural-log evidence. For keyword arguments see the
        `dynesty.run_nested()`_ documentation.

        Args:
            prior_function (callable, optional): the function to populated some
                prior distributions. Default is the broad uniform priors in
                uravu.relationship.Relationship.
            progress (bool, optional): Show tqdm progress for sampling.
                Default is `True`.

        .. _dynesty.run_nested(): https://dynesty.readthedocs.io/en/latest/api.html#dynesty.sampler.Sampler.run_nested
        """
        self.nested_sampling_results = sampling.nested_sampling(
            self, prior_function=prior_function, progress=progress, **kwargs
        )
        self.ln_evidence = ufloat(
            self.nested_sampling_results["logz"][-1],
            self.nested_sampling_results["logzerr"][-1],
        )



"""
Initialisation of uravu.
"""

import warnings
from pint import UnitRegistry, set_application_registry, Quantity

UREG = UnitRegistry()
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    Quantity([])
Q_ = UREG.Quantity
set_application_registry(UREG)


def __version__():
    """
    Return the current version of uravu.

    Returns:
        (str): Version number.
    """
    major = 0
    minor = 0
    micro = 1
    return "{}.{}.{}".format(major, minor, micro)



"""
The Distribution class allows the storage and analysis of probability
distributions.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import numpy as np
from scipy.stats import normaltest
from uncertainties import ufloat
from uravu import UREG


class Distribution:
    """
    The storage of probability distributions is fundamental to Bayesian
    inference, this class enables this. In addition to storage some basic
    analysis of the distribution is possible.

    Attributes:
        name (str): A name for the distribution.
        unit (pint.UnitRegistry(), optional): The unit of the values in the
            Distribution.
        samples (array_like): Samples in the distribution.
        ci_points (array_like, optional): The percentiles at which
            confidence intervals should be found. Default is
            `[2.5, 97.5]` (a 95 % confidence interval).
    """

    def __init__(
        self,
        samples,
        name="Distribution",
        ci_points=None,
        unit=UREG.dimensionless,
    ):
        """
        Args:
            samples (array_like): Sample for the distribution.
            name (str, optional): A name to identify the distribution.
                Default is `Distribution`.
            ci_points (array_like, optional): The percentiles at which
                confidence intervals should be found. Default is
                `[2.5, 97.5]` (a 95 % confidence interval).
            unit (pint.UnitRegistry(), optional) The unit for the
                distribution. Default is dimensionless.
        """
        self.name = name
        self.unit = unit
        self.samples = np.array([])
        if ci_points is None:
            self.ci_points = [2.5, 97.5]
        else:
            if len(ci_points) != 2:
                raise ValueError(
                    "The ci_points must be an array or tuple of length two."
                )
            self.ci_points = ci_points
        self.normal = False
        self.add_samples(np.array(samples))

    @property
    def mean(self):
        """
        Get the mean value and uncertainty. Will return ``None`` if
        distribution is not normal.

        Returns:
            (uncertainties.core.Variable or None): Mean value with uncertainty.
        """
        if self.normal:
            return ufloat(self.n, self.s)
        return None

    @property
    def size(self):
        """
        Get the number of samples in the distribution.

        Returns:
            (int): Number of samples.
        """
        return self.samples.size

    def check_normality(self):
        """
        Uses a Shapiro-Wilks statistical test to evaluate if samples are
        normally distributed.
        """
        alpha = 0.05
        if self.size <= 8:
            return False
        test_samples = self.samples
        if self.size > 500:
            test_samples = np.random.choice(self.samples, size=500)
        p_value = normaltest(test_samples)[1]
        if p_value > alpha:
            self.normal = True
        else:
            self.normal = False

    @property
    def n(self):
        """
        Get the median value of the distribution (for a normal distribution
        this is the same as the mean).

        Returns:
            (float): Median value.
        """
        return np.percentile(self.samples, [50])[0]

    @property
    def s(self):
        """
        Get the standard deviation of the distribution. For a non-normal
        distribution, this will return ``None``.

        Returns:
            (float, or None): Standard deviation of the distribution.
        """
        if self.normal:
            return np.std(self.samples)
        else:
            return None

    @property
    def con_int(self):
        """
        Get the extrema of the confidence intervals of the distribution.

        Returns:
            (array_like): The confidence interval values.
        """
        return np.percentile(self.samples, self.ci_points)

    def __repr__(self):
        """
        A custom representation, which is the same as the custom string
        representation.

        Returns:
            (str): String representation.
        """
        return self.__str__()

    def __str__(self):
        """
        A custom string.

        Returns:
            (str): Detailed string representation.
        """
        representation = "Distribution: {}\nSize: {}\n".format(
            self.name, self.size
        )
        representation += "Samples: "
        representation += "["
        representation += " ".join(
            ["{:.2e}".format(i) for i in self.samples[:2]]
        )
        representation += " ... "
        representation += " ".join(
            ["{:.2e}".format(i) for i in self.samples[-2:]]
        )
        representation += "]\n"
        representation += "Median: {:.2e}\n".format(self.n)
        if self.normal:
            representation += "Symetrical Error: {:.2e}\n".format(self.s)
        representation += "Confidence intervals: ["
        representation += " ".join(["{:.2e}".format(i) for i in self.con_int])
        representation += "]\n"
        representation += "Confidence interval points: ["
        representation += " ".join(["{}".format(i) for i in self.ci_points])
        representation += "]\n"
        if self.n is not None:
            representation += "Reporting Value: "
            if self.normal:
                representation += "{}\n".format(ufloat(self.n, self.s))
            else:
                representation += "{:.2e}+{:.2e}-{:.2e}\n".format(
                    self.n, self.con_int[1]-self.n, self.n-self.con_int[0]
                )
        representation += "Unit: {}\n".format(self.unit)
        representation += "Normal: {}\n".format(self.normal)
        return representation

    def add_samples(self, samples):
        """
        Add samples to the distribution and update values such as median and
        uncertainties as appropriate.

        Args:
            samples (array_like): Samples to be added to the distribution.
        """
        self.samples = np.append(self.samples, np.array(samples).flatten())
        self.check_normality()



"""
Tests for optimize module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import unittest
import numpy as np
from numpy.testing import assert_almost_equal
from uravu import optimize, utils, relationship


class TestOptimize(unittest.TestCase):
    """
    Unit tests for optimize module.
    """

    def test_ln_likelihood_a(self):
        """
        Test ln_likelihood function.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = np.ones(10)
        test_y_e = np.ones(10) * 0.1
        test_rel = relationship.Relationship(
            utils.straight_line, test_x, test_y, test_y_e
        )
        expected_lnl = -1724226.97414907
        actual_lnl = optimize.ln_likelihood(
            test_rel.variables, test_rel.function, test_rel.x, test_rel.y
        )
        assert_almost_equal(actual_lnl, expected_lnl)

    def test_ln_likelihood_b(self):
        """
        Test ln_likelihood function with no uncertainty.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = np.ones(10)
        test_rel = relationship.Relationship(
            utils.straight_line, test_x, test_y
        )
        expected_lnl = -45.21054955719477
        actual_lnl = optimize.ln_likelihood(
            test_rel.variables, test_rel.function, test_rel.x, test_rel.y, unaccounted_uncertainty=test_rel.unaccounted_uncertainty
        )
        assert_almost_equal(actual_lnl, expected_lnl)

    def test_negative_lnl_a(self):
        """
        Test negative_lnl function.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = np.ones(10)
        test_y_e = np.ones(10) * 0.1
        test_rel = relationship.Relationship(
            utils.straight_line, test_x, test_y, test_y_e
        )
        expected_negtive_lnl = 1724226.97414907
        actual_negative_lnl = optimize.negative_lnl(
            test_rel.variables, test_rel.function, test_rel.x, test_rel.y
        )
        assert_almost_equal(actual_negative_lnl, expected_negtive_lnl)

    def test_negative_lnl_b(self):
        """
        Test negative_lnl function with additional uncertainty.
        """
        test_x = np.linspace(0, 99, 10)
        test_y = np.ones(10)
        test_y_e = np.ones(10) * 0.1
        test_rel = relationship.Relationship(
            utils.straight_line,
            test_x,
            test_y,
            test_y_e,
            unaccounted_uncertainty=True,
        )
        expected_negtive_lnl = 45.21123241122563
        actual_negative_lnl = optimize.negative_lnl(
            test_rel.variables,
            test_rel.function,
            test_rel.x,
            test_rel.y,
            test_rel.unaccounted_uncertainty,
        )
        assert_almost_equal(actual_negative_lnl, expected_negtive_lnl)

    def test_max_lnlikelihood(self):
        """
        Test negative_lnl function.
        """
        test_x = np.linspace(0, 99, 100)
        test_y = np.linspace(1, 199, 100)
        test_y_e = test_y * 0.1
        test_rel = relationship.Relationship(
            utils.straight_line, test_x, test_y, test_y_e
        )
        actual_best_variables = optimize.max_ln_likelihood(test_rel)
        assert_almost_equal(actual_best_variables, np.array([2, 1]))



"""
The :mod:`sampling` module implements the use of a generalised
MCMC (using `emcee`_) and nested sampling (using `dynesty`_) for the
:class:`Relationship` objects.

.. _emcee: https://emcee.readthedocs.io/
.. _dynesty: https://dynesty.readthedocs.io/
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

import numpy as np
import emcee
import dynesty
from uravu import optimize
from uravu.distribution import Distribution


def mcmc(
    relationship,
    prior_function=None,
    walkers=100,
    n_samples=500,
    n_burn=500,
    progress=True,
):
    """
    Perform MCMC to get the probability distributions for the variables
    of the relationship.

    Args:
        relationship (uravu.relationship.Relationship): The relationship to
            determine the posteriors of.
        prior_function (callable, optional): The function to populated some
            prior distributions. Default is
            uravu.relationship.Relationship.prior.
        walkers (int, optional): Number of MCMC walkers. Default is `100`.
        n_samples (int, optional): Number of sample points. Default is
            `500`.
        n_burn (int, optional): Number of burn in samples. Default is
            `500`.
        progress (bool, optional): Show tqdm progress for sampling.
            Default is `True`.

    Returns:
        (dict): a dictionary with the Distrbutions as a list
            ('distributions'), the chain ('chain') and the samples as an
            ``array_like`` ('samples').
    """
    if prior_function is None:
        prior_function = relationship.prior

    initial_prior = np.zeros((walkers, len(relationship.variable_medians)))
    called_prior = prior_function()
    for i, p in enumerate(called_prior):
        initial_prior[:, i] = p.rvs(walkers)

    ndims = initial_prior.shape[1]

    sampler = emcee.EnsembleSampler(
        walkers,
        ndims,
        ln_probability,
        args=[
            relationship.function,
            relationship.abscissa,
            relationship.ordinate,
            relationship.unaccounted_uncertainty,
            called_prior,
        ],
    )
    sampler.run_mcmc(initial_prior, n_samples + n_burn, progress=progress)

    post_samples = sampler.get_chain(discard=n_burn).reshape((-1, ndims))

    distributions = []
    for i in range(ndims):
        distributions.append(
            Distribution(
                post_samples[:, i],
                name=relationship.variable_names[i],
                unit=relationship.variable_units[i],
            )
        )

    results = {
        "distributions": distributions,
        "chain": sampler.get_chain().reshape((-1, ndims)),
        "samples": post_samples,
    }
    return results


def ln_probability(
    variables, function, abscissa, ordinate, unaccounted_uncertainty, priors
):
    """
    Determine the natural log probability for a given set of variables, by
    summing the prior and likelihood.

    Args:
        variables (array_like): Variables for the function evaluation.
        function (callable): The function to be evaluated.
        abscissa (array_like): The abscissa values.
        ordinate (array_like): The ordinate values.
        unaccounted_uncertainty (bool): Should an unaccounted
            uncertainty parameter be considered.
        prior_function (callable, optional): The function to populated some
            prior distributions. Default is
            uravu.relationship.Relationship.prior.

    Returns:
        (float): Negative ln-probability between model and data, considering
            priors.
    """
    log_prior = 0
    for i, var in enumerate(variables):
        log_prior += priors[i].logpdf(var)
    if np.isneginf(log_prior):
        return -np.inf
    return log_prior + optimize.ln_likelihood(
        variables, function, abscissa, ordinate, unaccounted_uncertainty
    )


def nested_sampling(
    relationship, prior_function=None, progress=True, **kwargs
):
    """
    Perform the nested sampling in order to determine the Bayesian natural log
    evidence.

    Args:
        relationship (uravu.relationship.Relationship): The relationship to
            estimate the evidence for.
        prior_function (callable, optional): the function to populated some
            prior distributions. Default is the broad uniform priors in
            uravu.relationship.Relationship.
        progress (bool, optional): Show tqdm progress for sampling.
            Default is `True`.

    Keyword Args:
        See the `dynesty.run_nested()` documentation.

    Returns:
        (dict): The results from the dynesty run.
    """
    if prior_function is None:
        prior_function = relationship.prior
    priors = prior_function()
    sampler = dynesty.NestedSampler(
        optimize.ln_likelihood,
        nested_prior,
        len(relationship.variables),
        logl_args=[
            relationship.function,
            relationship.abscissa,
            relationship.ordinate,
            relationship.unaccounted_uncertainty,
        ],
        ptform_args=[priors],
    )

    sampler.run_nested(print_progress=progress, **kwargs)
    results = sampler.results
    return results


def nested_prior(array, priors):
    """
    Convert to dynesty prior style from at used within ``uravu``.

    Args:
        array (array_like): An array of random uniform numbers (0, 1].
            The shape of which is M x N, where M is the number of
            parameters and N is the number of walkers.
        prior_function (callable, optional): The function to populated some
            prior distributions. Default is
            uravu.relationship.Relationship.prior.

    Returns:
        (array_like): an array of random uniform numbers distributed in
            agreement with the priors.
    """
    broad = np.copy(array)
    for i, prior in enumerate(priors):
        broad[i] = prior.ppf(broad[i])
    return broad



"""
Figure parameters for uravu to help make nice plots.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey

from collections import OrderedDict
from matplotlib import rcParams, cycler

TABLEAU = OrderedDict(
    [
        ("blue", "#0173B2"),
        ("orange", "#DE8F05"),
        ("green", "#029E73"),
        ("red", "#D55E00"),
        ("purple", "#CC78BC"),
        ("brown", "#CA9161"),
        ("pink", "#FBAFE4"),
        ("grey", "#949494"),
        ("yellow", "#ECE133"),
        ("turquoise", "#56B4E9"),
    ]
)

FONTSIZE = 20
NEARLY_BLACK = "#161616"
LIGHT_GREY = "#F5F5F5"
WHITE = "#ffffff"

MASTER_FORMATTING = {
    "axes.formatter.limits": (-3, 3),
    "xtick.major.pad": 7,
    "ytick.major.pad": 7,
    "ytick.color": NEARLY_BLACK,
    "xtick.color": NEARLY_BLACK,
    "axes.labelcolor": NEARLY_BLACK,
    "axes.spines.bottom": False,
    "axes.spines.left": False,
    "axes.spines.right": False,
    "axes.spines.top": False,
    "axes.axisbelow": True,
    "legend.frameon": False,
    "pdf.fonttype": 42,
    "ps.fonttype": 42,
    "mathtext.fontset": "custom",
    "font.size": FONTSIZE,
    "font.family": "serif",
    # 'font.serif' : 'Source Serif Pro',
    "text.usetex": True,
    "savefig.bbox": "tight",
    "axes.facecolor": LIGHT_GREY,
    "axes.labelpad": 10.0,
    "axes.labelsize": FONTSIZE * 0.8,
    "axes.titlepad": 30,
    "axes.titlesize": FONTSIZE,
    "axes.grid": False,
    "grid.color": WHITE,
    "lines.markersize": 7.0,
    "lines.scale_dashes": False,
    "xtick.labelsize": FONTSIZE * 0.8,
    "ytick.labelsize": FONTSIZE * 0.8,
    "legend.fontsize": FONTSIZE * 0.8,
    "lines.linewidth": 4,
}

for k, v in MASTER_FORMATTING.items():
    rcParams[k] = v

COLOR_CYCLE = TABLEAU.values()

rcParams["axes.prop_cycle"] = cycler(color=COLOR_CYCLE)
